
#ifndef H_PLAINUTF8_INCLUDED
 #define H_PLAINUTF8_INCLUDED

/*
 *      OVERVIEW
 *
 * This header file provides equivalents for all the standard library
 * functions in <string.h> and the non-standard <strings.h> which are
 * applicable to programs that use plain C strings (zero-terminated
 * arrays of "char") to store UTF-8 data. It also provides a handful
 * of standard library functions from <wchar.h>.
 *
 * The idea is that the program can be written entirely in terms of plain C
 * strings, completely avoiding the arcane problems of "wchar_t". Therefore,
 * this library does not provide most of the functionality of <wchar.h>.
 * Only the functions to convert between plain C strings and arrays of
 * "wchar_t" --- mbstowcs, wcstombs, and their relatives --- are provided.
 * Terminal output functions such as fputwc are not provided, because the
 * idea is that all input and output can and should be done via plain C
 * strings, using the standard library's getc/putc/scanf/printf
 * implementation.
 *
 *
 *      FEATURES CONTROLLABLE VIA MACROS
 *
 * Some features of this library are controllable via #defined macros.
 *
 * PLAINUTF8_PROVIDE_REDUNDANT_UTF8_PREFIXES
 * Certain string functions work just as well with UTF-8 strings as
 * they do with ASCII strings. For example, strcpy. This header file
 * does not provide a declaration for utf8_strcpy by default.
 * #define PLAINUTF8_PROVIDE_REDUNDANT_UTF8_PREFIXES 1
 * will instruct this header file to #define utf8_strcpy strcpy,
 * and so on for the other <string.h> functions unaffected by the
 * transition to UTF-8.
 *
 * PLAINUTF8_DROPIN_REPLACEMENTS
 * PLAINUTF8_DROPIN_ALLOW_STRCHR
 * PLAINUTF8_DROPIN_ALLOW_ALL
 * To quickly UTF-8-ize an existing program, it may suffice
 * to #define strstr as a macro expanding to utf8_strstr, and so on.
 * #define PLAINUTF8_DROPIN_REPLACEMENTS 1
 * will instruct this header file to #define strstr utf8_strstr,
 * and so on for the other <string.h> functions affected by the
 * transition to UTF-8. This option is essentially the inverse of
 * PLAINUTF8_PROVIDE_REDUNDANT_UTF8_PREFIXES, yet the two options
 * are orthogonal and can be used together or separately.
 * As a safety measure, PLAINUTF8_DROPIN_REPLACEMENTS will by default
 * forbid the use of a handful of "dangerous" functions (such as strncat
 * and strchr) by #defining them in such a way that any use of these
 * functions will cause a compile-time syntax error.
 * #define PLAINUTF8_DROPIN_ALLOW_STRCHR 1
 * will instruct this header file not to forbid strchr or strrchr.
 * #define PLAINUTF8_DROPIN_ALLOW_ALL 1
 * will instruct this header file not to forbid any functions,
 * even "dangerous" ones. (Neither of these last two options imply
 * PLAINUTF8_DROPIN_REPLACMENTS; that must be #defined separately.)
 *
 * PLAINUTF8_PROVIDE_MBTOWC
 * When this macro is #defined non-zero, this header file will
 * provide declarations for utf8_mbtowc and a handful of related
 * functions. These functions are unaffected by
 * PLAINUTF8_DROPIN_REPLACEMENTS.
 *
 *
 *      CAVEATS, LOOPHOLES AND SECURITY IMPLICATIONS 
 *
 * Note that certain features of scanf and printf will not work properly
 * with UTF-8 strings. Namely:
 *     scanf("%[u]")         where "u" contains non-ASCII characters
 *     printf("%8s", us)     where us contains non-ASCII characters
 *     printf("%*s", i, us)  where us contains non-ASCII characters
 *     printf("%.8s", us)    where us contains non-ASCII characters
 *     printf("%.*s", i, us) where us contains non-ASCII characters
 * This library does not attempt to address these limitations.
 *
 * Certain string functions simply do not make sense in a UTF-8 environment,
 * because they take arguments of type "char" which are unable to represent
 * the full range of Unicode. Generally speaking, it is possible to rewrite
 * expressions using one of these library functions to use a similar function
 * that takes a plain C string instead of a single "char". For example:
 *     strchr("s", 'c')     should be rewritten as  utf8_strstr("s", "c")
 *     strrchr("s", 'c')    should be rewritten as  utf8_strrstr("s", "c")
 *
 * The strncpy and strncat functions are typically used to guard against
 * buffer overflows by passing the size in bytes of a character buffer;
 * in contrast, this library's utf8_strncpy() copies up to "n" Unicode
 * characters, which may amount to as many as "6*n" bytes. It would be
 * irresponsible to replace a program's uses of strncpy() with
 * utf8_strncpy(), without first checking the surrounding code very
 * carefully for potential buffer overflows. For this reason, this header
 * file will not #define strncpy, strncat, or strncmp, even when
 * PLAINUTF8_DROPIN_REPLACEMENTS is #defined.
 *
 * The functions utf8_strtok, utf8_strtok_r, and utf8_strsep modify
 * their string arguments by overwriting a delimiter character. This is
 * done by writing '\0' into the first byte of the delimiter character's
 * byte sequence, and leaving the other bytes untouched. This is efficient
 * and generally harmless, but can result in garbage bytes between the
 * terminating '\0' of one token and the first character of the next.
 * For example, strtok("Erd\xC5\x91s", "\xC5\91") returns "Erd\0\x91s",
 * and a second call to strtok(NULL, "\xC5\x91") returns the final "s",
 * having already skipped over the now-garbage byte "\x91". However,
 * the extra "\x91" might pose a problem for hand-written C code that
 * expects to be able to parse through a buffer tokenized with strtok
 * simply by skipping over any embedded '\0' bytes it encounters.
 *
 *
 *      UNSPECIFIED AND UNDEFINED BEHAVIOR
 *
 * All the functions defined in this header file have undefined behavior
 * under any circumstances under which their standard library counterparts
 * would have undefined behavior. For example, passing NULL to most of
 * these functions yields undefined behavior.
 * 
 * All the functions defined in this header file have unspecified
 * behavior when presented with invalid input (as defined by RFC 3629).
 * However, under those circumstances they will not exhibit undefined
 * behavior, and they will terminate. For example, it is unspecified
 * what value will be returned by utf8_strlen("\xC0\x80"), but such a
 * call will return *some* value, and it will not dereference any memory
 * location outside the given zero-terminated string.
 */


/* The implementation of "plainutf8.h" begins here. */

typedef char utf8_char;

#if PLAINUTF8_DROPIN_REPLACEMENTS
 /* If we do not #include <string.h> now, the user might try
  * to #include it later, leading to syntax errors in the standard
  * library's function prototypes for strncpy and friends.
  * Since we do #include it, such errors will probably be prevented
  * by the implementation's inclusion guards. */
 #include <string.h>
#endif /* PLAINUTF8_DROPIN_REPLACEMENTS */

#if PLAINUTF8_PROVIDE_MBTOWC
 /* If we are to provide a prototype for utf8_mbtowc, we must
  * get a definition for the type "wchar_t".
  */
 #include <stddef.h>
#endif /* PLAINUTF8_PROVIDE_MBTOWC */


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* These functions behave similarly to their C counterparts, but
 * have been extended to deal with UTF-8 strings. */
size_t       utf8_strcspn(const utf8_char *s, const utf8_char *reject);
utf8_char   *utf8_strfry(utf8_char *string);
size_t       utf8_strlen(const utf8_char *s);
utf8_char   *utf8_strpbrk(const utf8_char *s, const utf8_char *accept);
utf8_char   *utf8_strsep(utf8_char **stringp, const utf8_char *delim);
size_t       utf8_strspn(const utf8_char *s, const utf8_char *accept);
utf8_char   *utf8_strtok(utf8_char *s, const utf8_char *delim);
utf8_char   *utf8_strtok_r(utf8_char *s, const utf8_char *delim, utf8_char **saveptr);

/* These functions' parameter "n" does not refer to a count of bytes.
 * It refers to a count of UTF-8 characters, like the count returned
 * by utf8_strlen() or utf8_strspn(). This has security implications
 * related to buffer overflow.
 */
utf8_char   *utf8_strncat(utf8_char *dest, const utf8_char *src, size_t n);
int          utf8_strncmp(const utf8_char *s1, const utf8_char *s2, size_t n);
utf8_char   *utf8_strncpy(utf8_char *dest, const utf8_char *src, size_t n);


#if PLAINUTF8_PROVIDE_MBTOWC
 /* Note that the parameters "n" of these functions do not refer to
  * counts of UTF-8 characters. These functions are literally
  * re-implementations of the standard C99 library functions, with
  * the sole distinguishing feature being that they specifically
  * implement C99's "multibyte strings" using UTF-8 encoding.
  *
  * These functions are provided merely as a convenience for users who
  * require random access to their Unicode strings. The entire idea
  * of this library is that arrays of "wchar_t" should never be used
  * for input or output.
  */
 int         utf8_mbtowc(wchar_t *pwc, const utf8_char *s, size_t n);
 int         utf8_wctomb(utf8_char *s, wchar_t wc);
 size_t      utf8_wcstombs(utf8_char *dest, const wchar_t *src, size_t n);
 size_t      utf8_mbstowcs(wchar_t *dest, const utf8_char *src, size_t n);
#endif /* PLAINUTF8_PROVIDE_MBTOWC */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#if PLAINUTF8_DROPIN_REPLACEMENTS
 #undef strcspn
 #define strcspn utf8_strcspn
 #undef strfry
 #define strfry  utf8_strfry
 #undef strlen
 #define strlen  utf8_strlen
 #undef strpbrk
 #define strpbrk utf8_strpbrk
 #undef strsep
 #define strsep  utf8_strsep
 #undef strspn
 #define strspn  utf8_strspn
 #undef strtok
 #define strtok  utf8_strtok
 #if !(PLAINUTF8_ALLOW_STRCHR || PLAINUTF8_ALLOW_ALL)
  #undef strchr
  #define strchr       ..PLAINUTF8_DROPIN_REPLACEMENTS disables strchr..
  #undef strrchr
  #define strrchr      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strrchr..
 #endif
 #if !(PLAINUTF8_ALLOW_ALL)
  #undef strcasecmp
  #define strcasecmp   ..PLAINUTF8_DROPIN_REPLACEMENTS disables strcasecmp..
  #undef strcoll
  #define strcoll      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strcoll..
  #undef strncasecmp
  #define strncasecmp  ..PLAINUTF8_DROPIN_REPLACEMENTS disables strncasecmp..
  #undef strncat
  #define strncat      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strncat..
  #undef strncpy
  #define strncpy      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strncpy..
  #undef strncmp
  #define strncmp      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strncmp..
  #undef strxfrm
  #define strxfrm      ..PLAINUTF8_DROPIN_REPLACEMENTS disables strxfrm..
 #endif
#endif /* PLAINUTF8_DROPIN_REPLACEMENTS */

#if PLAINUTF8_PROVIDE_REDUNDANT_UTF8_PREFIXES
 #define utf8_strcat  strcat
 #define utf8_strcmp  strcmp
 #define utf8_strcpy  strcpy
 #define utf8_strdup  strdup
 #define utf8_strrstr strrstr
 #define utf8_strstr  strstr
#endif /* PLAINUTF8_PROVIDE_REDUNDANT_UTF8_PREFIXES */

#endif /* H_PLAINUTF8_INCLUDED */
