
#include "uassert.h"

/*
 * DESCRIPTION
 * If dest is not a NULL pointer, the mbstowcs() function converts the
 * multibyte string src to a wide-character string starting at dest. At
 * most n wide characters are written to dest. The conversion starts in
 * the initial state. The conversion can stop for three reasons:
 *
 * 1. An invalid multibyte sequence has been encountered. In this case
 *    (size_t) -1 is returned.
 *
 * 2. n non-L'\0' wide characters have been stored at dest. In this case
 *    the number of wide characters written to dest is returned, but the
 *    shift state at this point is lost.
 *
 * 3. The multibyte string has been completely converted, including the
 *    terminating '\0'. In this case the number of wide characters written
 *    to dest, excluding the terminating L'\0' character, is returned.
 *
 * The programmer must ensure that there is room for at least n wide
 * characters at dest.
 *
 * If dest is NULL, n is ignored, and the conversion proceeds as above,
 * except that the converted wide characters are not written out to memory,
 * and that no length limit exists.
 *
 * In order to avoid the case 2 above, the programmer should make sure n
 * is greater than or equal to mbstowcs(NULL,src,0)+1.
 *
 * RETURN VALUE
 * The mbstowcs() function returns the number of wide characters that make
 * up the converted part of the wide-character string, not including the
 * terminating null wide character. If an invalid multibyte sequence was
 * encountered, (size_t) -1 is returned.
 */

size_t utf8_mbstowcs(wchar_t *dest, const utf8_char *src, size_t n)
{
    const int dest_not_null = (dest != NULL);
    size_t count = 0;

    assert(src != NULL);
    /* Note that we're sort of assuming sizeof(wchar_t)==4, but
     * nevertheless nothing terribly bad will happen if it's not.
     */
    
    while (count < n) {
        uint32 ch32;
        if (*src == '\0') {
            /* Case 3. */
            if (dest_not_null)
              *dest++ = '\0';
            return count;
        }
        ch32 = plainutf8_scan(src, &src);
        /* Generally we will treat even invalid bytestreams as valid; for
         * example, mbstowcs(dest, "\xD2", 2) will store 0xD2 in dest[0]
         * and 0x00 in dest[1]. But we make one important exception! If
         * ch32 is '\0' at this point, we will reject the bytestream as
         * invalid. For example, we would reject mbstowcs(dest, "\xC080", 1).
         * This is because the standard requires us never to write a
         * wchar_t value of L'\0' into dest unless it came from a '\0'
         * in src. */
        if (ch32 == 0) {
            /* Case 1. */
            return (size_t) -1;
        }
        if (dest_not_null)
          *dest++ = ch32;
        ++count;
    }
    /* Case 2. */
    return count;
}

