
#include "uassert.h"

/*
 * DESCRIPTION
 * The main case for this function is when s is not NULL and pwc is not
 * NULL. In this case, the mbtowc() function inspects at most n bytes of
 * the multibyte string starting at s, extracts the next complete multi-
 * byte character, converts it to a wide character and stores it at *pwc.
 * It updates an internal shift state known only to the mbtowc function.
 * If s does not point to a '\0' byte, it returns the number of bytes that
 * were consumed from s, otherwise it returns 0.
 *
 * If the n bytes starting at s do not contain a complete multibyte
 * character, or if they contain an invalid multibyte sequence, mbtowc()
 * returns -1. This can happen even if n >= MB_CUR_MAX, if the multibyte
 * string contains redundant shift sequences.
 *
 * A different case is when s is not NULL but pwc is NULL. In this case
 * the mbtowc() function behaves as above, except that it does not store
 * the converted wide character in memory.
 *
 * A third case is when s is NULL. In this case, pwc and n are ignored.
 * The mbtowc() function resets the shift state, only known to this
 * function, to the initial state, and returns nonzero if the encoding has
 * nontrivial shift state, or zero if the encoding is stateless.
 *
 * RETURN VALUE
 * If s is not NULL, the mbtowc() function returns the number of consumed
 * bytes starting at s, or 0 if s points to a null byte, or -1 upon failure.
 *
 * If s is NULL, the mbtowc() function returns nonzero if the encoding has
 * nontrivial shift state, or zero if the encoding is stateless.
 */

#define MALFORMED_INPUT goto finish

#define GET_BYTE                                 \
    if (i >= n) return -1;                       \
    ch = *us++;                                  \
    if (ch < 0x80 || 0xBF < ch) MALFORMED_INPUT; \
    result = (result << 6) | (ch & ~0x80);       \
    ++i;

int utf8_mbtowc(wchar_t *pwc, const utf8_char *s, size_t n)
{
    const unsigned char *us = (const unsigned char *)s;
    unsigned char ch;
    uint32 result = 0;
    size_t i;
    
    if (s == NULL) {
        /* "A third case." Reset the (non-existent) hidden state. */
        return 0;
    } else if (n == 0) {
        return -1;
    }
    
    /* Otherwise, we should scan "s" for a UTF-8 character. If
     * we find one, store it in "pwc" and return the number of
     * bytes consumed. If we see an invalid encoding, we are
     * supposed to return -1; but we are as lenient as possible.
     * If the first byte of "s" is invalid, then scan that one byte.
     * If the first byte of "s" indicates a multi-byte sequence
     * but one of the later bytes is invalid, then scan up to the
     * offending byte and leave the offending byte unscanned.
     * If the first byte of "s" indicates a multi-byte sequence
     * but "n" is too short for the sequence, then (and only then)
     * return -1 without scanning any input.
     */
    assert(us != NULL);

    result = *us++;
    i = 1;
    if (result <= 0x7F) {
        /* continue */
    } else if (result <= 0xC1) {
        MALFORMED_INPUT;
    } else if (result <= 0xDF) {
        /* Two-byte sequence. */
        result &= ~0xC0;
        GET_BYTE
    } else if (result <= 0xEF) {
        /* Three-byte sequence. */
        result &= ~0xE0;
        GET_BYTE
        GET_BYTE
    } else if (result <= 0xF7) {
        /* Four-byte sequence. */
        result &= ~0xF0;
        GET_BYTE GET_BYTE
        GET_BYTE
    } else if (result <= 0xFB) {
        /* Five-byte sequence. */
        result &= ~0xF8;
        GET_BYTE GET_BYTE
        GET_BYTE GET_BYTE
    } else if (result <= 0xFD) {
        /* Six-byte sequence. */
        result &= ~0xFC;
        GET_BYTE GET_BYTE
        GET_BYTE GET_BYTE
        GET_BYTE
    } else {
        MALFORMED_INPUT;
    }
  finish:
    if (pwc != NULL) *pwc = result;
    return i;
}

