
#include "uassert.h"
#include <stdlib.h>  /* rand */

extern size_t utf8_strlen(const utf8_char *s);

static int is_malformed(const utf8_char *s);
static void triswap(utf8_char *a, utf8_char *aend,
                    utf8_char *b, utf8_char *bend);
static void reverse(utf8_char *p, size_t len);

/*
 * DESCRIPTION
 * The strfry() function randomizes the contents of string by using
 * rand(3) to randomly swap characters in the string. The result is an
 * anagram of string.
 * RETURN VALUE
 * The strfry() functions returns a pointer to the randomized string.
 */
utf8_char *utf8_strfry(utf8_char *string)
{
    /* This function is of course a joke on GNU's part.
     * The "randomizing" algorithm is bad in at least two ways:
     * [A], it takes rand() modulo another (non-power-of-two) number,
     * which introduces bias towards the low numbers; and [B],
     * it shuffles an array by swapping string[i] with
     * string[rand(0..len-1)], which introduces a bias that would not
     * be present if string[i] were instead swapped with
     * string[rand(i..len-1)]. Those (intentional?) algorithmic
     * blunders are faithfully reproduced here.
     * In addition, whereas strfry is O(n) overall, the difficulty
     * of doing both random access and "swapping characters" on a
     * UTF-8 string makes utf8_strfry O(n^2)!
     */
    size_t len;
    size_t i;
    utf8_char *ptri;

    assert(string != NULL);

    /* If the given string is malformed in any way, we reject it;
     * this is the simplest way to ensure that all this byte-swapping
     * never results in our changing the length of the string. */
    if (is_malformed(string))
      return string;

    len = utf8_strlen(string);
    ptri = string;
    for (i=0; i < len; ++i) {
        utf8_char *endptri;
        size_t j = rand() % len;
        (void)plainutf8_scan(ptri, &endptri);
        if (i != j) {
            utf8_char *ptrj, *endptrj;
            for (ptrj = string; j != 0; --j) {
                assert(*ptrj != '\0');
                (void)plainutf8_scan(ptrj, &ptrj);
            }
            (void)plainutf8_scan(ptrj, &endptrj);
            triswap(ptri, endptri, ptrj, endptrj);
	    if (ptri < ptrj) {
		endptri = ptri + (endptrj - ptrj);
	    }
        }
	ptri = endptri;
    }
    
    return string;
}

/* Given pointers a and b into the string, swap the characters
 * they point to, like this:
 *
 *    (a..aend) (aend..b) (b..bend)
 * becomes
 *    (b..bend) (aend..b) (a..aend)
 *
 * We do this by reversing the whole stretch in place, and then
 * re-reversing each of the three segments in place.
 */
static void triswap(utf8_char *a, utf8_char *aend,
                    utf8_char *b, utf8_char *bend)
{
    if (a > b) {
        triswap(b, bend, a, aend);
        return;
    } else {
        size_t len1, len2, len3;
        assert(a < aend);
        assert(aend <= b);
        assert(b < bend);
        len1 = aend - a;
        len2 = b - aend;
        len3 = bend - b;
        reverse(a, len1+len2+len3);
        reverse(a, len3);
        reverse(a+len3, len2);
        reverse(a+len3+len2, len1);
        return;
    }
}

static void reverse(utf8_char *p, size_t len)
{
    size_t i;
    for (i=0; i+1 < len; ++i, --len) {
        const utf8_char temp = p[i];
        p[i] = p[len-1];
        p[len-1] = temp;
    }
}



#define MALFORMED_INPUT return 1

#define GET_BYTE                                 \
    ch = *us++;                                  \
    if (ch < 0x80 || 0xBF < ch) MALFORMED_INPUT;

static int is_malformed(const utf8_char *s)
{
    const unsigned char *us = (const unsigned char *)s;
    unsigned char ch;
    assert(us != NULL);

    for (ch = *us++; ch != '\0'; ch = *us++) {
        if (ch <= 0x7F) {
            /* continue */
        } else if (ch <= 0xC1) {
            MALFORMED_INPUT;
        } else if (ch <= 0xDF) {
            /* Two-byte sequence. */
            GET_BYTE
        } else if (ch <= 0xEF) {
            /* Three-byte sequence. */
            GET_BYTE
            GET_BYTE
        } else if (ch <= 0xF7) {
            /* Four-byte sequence. */
            GET_BYTE GET_BYTE
            GET_BYTE
        } else if (ch <= 0xFB) {
            /* Five-byte sequence. */
            GET_BYTE GET_BYTE
            GET_BYTE GET_BYTE
        } else if (ch <= 0xFB) {
            /* Six-byte sequence. */
            GET_BYTE GET_BYTE
            GET_BYTE GET_BYTE
            GET_BYTE
        } else {
            MALFORMED_INPUT;
        }
    }
    /* The input was not malformed. */
    return 0;
}

