
#include "uassert.h"

/*
 * DESCRIPTION
 * The strtok() function parses a string into a sequence of tokens. On
 * the first call to strtok() the string to be parsed should be specified
 * in str. In each subsequent call that should parse the same string, str
 * should be NULL.
 *
 * The delim argument specifies a set of characters that delimit the
 * tokens in the parsed string. The caller may specify different strings
 * in delim in successive calls that parse the same string.
 *
 * Each call to strtok() returns a pointer to a null-terminated string
 * containing the next token. This string does not include the delimiting
 * character. If no more tokens are found, strtok() returns NULL.
 *
 * A sequence of two or more contiguous delimiter characters in the parsed
 * string is considered to be a single delimiter. Delimiter characters at
 * the start or end of the string are ignored. Put another way: the
 * tokens returned by strtok() are always nonempty strings.
 *
 * The strtok_r() function is a reentrant version of strtok(). The saveptr
 * argument is a pointer to a char * variable that is used internally by
 * strtok_r() in order to maintain context between successive calls that
 * parse the same string.
 *
 * On the first call to strtok_r(), str should point to the string to be
 * parsed, and the value of saveptr is ignored. In subsequent calls, str
 * should be NULL, and saveptr should be unchanged since the previous
 * call.
 *
 * Different strings may be parsed concurrently using sequences of calls
 * to strtok_r() that specify different saveptr arguments.
 *
 * RETURN VALUE
 * The strtok() and strtok_r() functions return a pointer to the next
 * token, or NULL if there are no more tokens.
 */

utf8_char *utf8_strtok_r(utf8_char *s, const utf8_char *delim,
                         utf8_char **saveptr)
{
    utf8_char *next_s;
    utf8_char *start_of_token;

    assert(delim != NULL);
    assert(saveptr != NULL);

    if (s == NULL) {
        /* After each call to utf8_strtok_r, we update *saveptr
         * to point at the next character after the delimiter we
         * found (and set to zero). */
        s = *saveptr;
    }
    
    /* Skip over any leading delimiter characters. */
    while (*s != '\0') {
        const uint32 ch32 = plainutf8_scan(s, &next_s);
        const utf8_char *needle;
        for (needle = delim; *needle != '\0'; ) {
            if (ch32 == plainutf8_scan(delim, &delim))
              goto found_leading_delimiter;
        }
        /* This "ch32" was not found in "delim". */
        goto found_token;
      found_leading_delimiter:
        s = next_s;
    }

    if (*s == '\0') {
        /* The string consisted entirely of leading delimiters,
         * or it was empty. */
        *saveptr = s;
        return NULL;
    }

  found_token:
    assert(*s != '\0');
    start_of_token = s;
    s = next_s;

    /* Look for a trailing delimiter or '\0'. */
    while (*s != '\0') {
        const uint32 ch32 = plainutf8_scan(s, &next_s);
        const utf8_char *needle;
        for (needle = delim; *needle != '\0'; ) {
            if (ch32 == plainutf8_scan(delim, &delim)) {
              *s = '\0';
              goto found_trailing_delimiter_or_zero;
            }
        }
        /* This "ch32" was not found in "delim". */
        s = next_s;
    }

  found_trailing_delimiter_or_zero:    
    assert(*s == '\0');
    *saveptr = next_s;
    return start_of_token;
}

