
#include "uassert.h"

/*
 * DESCRIPTION
 * If s is not NULL, the wctomb() function converts the wide character wc
 * to its multibyte representation and stores it at the beginning of the
 * character array pointed to by s.  It updates the shift state, which is
 * stored in a static anonymous variable only known to the wctomb function,
 * and returns the length of said multibyte representation, that is,
 * the number of bytes written at s.
 *
 * The programmer must ensure that there is room for at least MB_CUR_MAX
 * bytes at s.
 *
 * If s is NULL, the wctomb() function resets the shift state, known only
 * to this function, to the initial state, and returns nonzero if the
 * encoding has nontrivial shift state, or zero if the encoding is stateless.
 *
 * RETURN VALUE
 * If s is not NULL, the wctomb() function returns the number of bytes
 * that have been written to the byte array at s. If wc cannot be represented
 * as a multibyte sequence (according to the current locale), -1 is returned.
 */

int utf8_wctomb(utf8_char *s, wchar_t wc)
{
    if (s == NULL) {
        /* Reset the (non-existent) hidden state. */
        return 0;
    }
    
    if (wc <= 0x7F) {
        s[0] = wc;
        return 1;
    } else if (wc <= 0x7FF) {
        s[0] = 0xC0 | ((wc >> 6) & 0x1F);
        s[1] = 0x80 | ((wc >> 0) & 0x3F);
        return 2;
    } else if (wc <= 0xFFFF) {
        s[0] = 0xE0 | ((wc >> 12) & 0x0F);
        s[1] = 0x80 | ((wc >> 6) & 0x3F);
        s[2] = 0x80 | ((wc >> 0) & 0x3F);
        return 3;
    } else if (wc <= 0x1FFFFF) {
        s[0] = 0xF0 | ((wc >> 18) & 0x07);
        s[1] = 0x80 | ((wc >> 12) & 0x3F);
        s[2] = 0x80 | ((wc >> 6) & 0x3F);
        s[3] = 0x80 | ((wc >> 0) & 0x3F);
        return 4;
    } else if (wc <= 0x7FFFFFF) {
        s[0] = 0xF8 | ((wc >> 24) & 0x03);
        s[1] = 0x80 | ((wc >> 18) & 0x3F);
        s[2] = 0x80 | ((wc >> 12) & 0x3F);
        s[3] = 0x80 | ((wc >> 6) & 0x3F);
        s[4] = 0x80 | ((wc >> 0) & 0x3F);
        return 5;
    } else if (wc <= 0x7FFFFFFF) {
        s[0] = 0xFC | ((wc >> 30) & 0x01);
        s[1] = 0x80 | ((wc >> 24) & 0x3F);
        s[2] = 0x80 | ((wc >> 18) & 0x3F);
        s[3] = 0x80 | ((wc >> 12) & 0x3F);
        s[4] = 0x80 | ((wc >> 6) & 0x3F);
        s[5] = 0x80 | ((wc >> 0) & 0x3F);
        return 6;
    } else {
        return -1;
    }
}

